from django import forms
from elasticsearch import Elasticsearch
from elasticsearch_dsl import Search
from elasticsearch_dsl.result import Response

from . import settings as es_settings


class ESSearchForm(forms.Form):

    def __init__(self, *args, **kwargs):
        self.query_params = kwargs.pop('query_params', {}).copy()
        self.es = kwargs.pop('es', None)

        super(ESSearchForm, self).__init__(*args, **kwargs)

    def get_index(self):
        # return the ES index name (or multiple comma-separated) you want to
        # target, or '' if you don't want to target an index
        return ''

    def get_type(self):
        # return the ES type name (or multiple comma-separated) you want to
        # target, or '' if you don't want to target a type
        return ''

    def prepare_query(self):
        raise NotImplementedError

    def search(self, page=1, page_size=20):
        esp = ESSearchProcessor(self.es)
        esp.add_search(self, page, page_size)
        responses = esp.search()

        # there will only be a single response from a ESSearchForm
        return responses[0]


class ESSearchProcessor(object):

    def __init__(self, es=None):
        self.es = es or Elasticsearch(es_settings.ELASTICSEARCH_SERVER)
        self.bulk_search_data = []
        self.page_ranges = []

    def reset(self):
        self.bulk_search_data = []
        self.page_ranges = []

    def add_search(self, query, page=1, page_size=20, index='', doc_type='', query_params={}):
        if isinstance(query, ESSearchForm):
            form = query
            index = index or form.get_index()
            doc_type = doc_type or form.get_type()

            qp = form.query_params.copy()
            qp.update(query_params)
            query_params = qp

            query = form.prepare_query()
        elif isinstance(query, Search):
            dsl_search = query
            index = index or dsl_search._index
            doc_type = doc_type or dsl_search._doc_type

            qp = dsl_search._params.copy()
            qp.update(query_params)
            query_params = qp

            query = dsl_search.to_dict()
        elif isinstance(query, dict):
            pass
        else:
            # we don't support any other type of object
            return

        query['from'] = (page - 1) * page_size
        query['size'] = page_size

        # save these here so we can attach the info the the responses below
        self.page_ranges.append((page, page_size))

        data = query_params.copy()
        if index:
            data['index'] = index
        if doc_type:
            data['type'] = doc_type

        self.bulk_search_data.append(data)
        self.bulk_search_data.append(query)

    def search(self):
        responses = []

        if self.bulk_search_data:
            data = self.es.msearch(self.bulk_search_data)
            if data:
                for i, tmp in enumerate(data.get('responses', [])):
                    response = Response(tmp)

                    # hack the from & size values into response.hits so the user
                    # can do some (fake) Django pagination if desired
                    hits = response.get('hits', {})
                    hits['from'], hits['size'] = self.page_ranges[i]

                    responses.append(response)

        self.reset()

        return responses



from django.conf import settings


if getattr(settings, 'IS_TEST', False):
    from django.db import models
    from django.db.models.signals import post_save, pre_delete

    from .mixins import ElasticsearchIndexMixin

    class Blog(models.Model):
        name = models.CharField(max_length=50)
        description = models.TextField()

    class BlogPost(models.Model, ElasticsearchIndexMixin):
        blog = models.ForeignKey(Blog)
        slug = models.SlugField()
        title = models.CharField(max_length=50)
        body = models.TextField()
        created_at = models.DateTimeField(auto_now_add=True)

        @classmethod
        def get_bulk_index_limit(cls):
            return 2

        @classmethod
        def get_queryset(cls):
            return BlogPost.objects.all().select_related('blog')

        @classmethod
        def get_index_name(cls):
            return 'blog'

        @classmethod
        def get_type_name(cls):
            return 'posts'

        @classmethod
        def get_type_mapping(cls):
            return {
                "properties": {
                    "created_at": {
                        "type": "date",
                        "format": "dateOptionalTime"
                    },
                    "title": {
                        "type": "string"
                    },
                    "body": {
                        "type": "string"
                    },
                    "slug": {
                        "type": "string"
                    },
                    "blog": {
                        "properties": {
                            "id": {
                                "type": "long"
                            },
                            "name": {
                                "type": "string"
                            },
                            "description": {
                                "type": "string"
                            }
                        }
                    }
                }
            }

        @classmethod
        def get_document(cls, obj):
            return {
                'created_at': obj.created_at,
                'title': obj.title,
                'body': obj.body,
                'slug': obj.slug,
                'blog': {
                    'id': obj.blog.id,
                    'name': obj.blog.name,
                    'description': obj.blog.description,
                }
            }

        @classmethod
        def should_index(cls, obj):
            return obj.slug != 'DO-NOT-INDEX'

    post_save.connect(BlogPost.save_handler, sender=BlogPost)
    pre_delete.connect(BlogPost.delete_handler, sender=BlogPost)



from elasticsearch import Elasticsearch, TransportError

from . import settings as es_settings
from .exceptions import MissingObjectError
from .utils import queryset_iterator


class ElasticsearchIndexMixin(object):

    @classmethod
    def get_es(cls):
        if not hasattr(cls, '_es'):
            cls._es = Elasticsearch(**cls.get_es_connection_settings())
        return cls._es

    @classmethod
    def get_es_connection_settings(cls):
        return es_settings.ELASTICSEARCH_CONNECTION_PARAMS

    @classmethod
    def get_index_name(cls):
        raise NotImplementedError

    @classmethod
    def get_type_name(cls):
        raise NotImplementedError

    @classmethod
    def get_document(cls, obj):
        raise NotImplementedError

    @classmethod
    def get_document_id(cls, obj):
        if not obj:
            raise MissingObjectError
        return obj.pk

    @classmethod
    def get_request_params(cls, obj):
        return {}

    @classmethod
    def get_type_mapping(cls):
        return {}

    @classmethod
    def get_queryset(cls):
        raise NotImplementedError

    @classmethod
    def get_bulk_index_limit(cls):
        return 100

    @classmethod
    def get_query_limit(cls):
        return 100

    @classmethod
    def should_index(cls, obj):
        return True

    @classmethod
    def bulk_index(cls, es=None, index_name='', queryset=None):
        es = es or cls.get_es()

        tmp = []

        if queryset is None:
            queryset = cls.get_queryset()

        # this requires that `get_queryset` is implemented
        for i, obj in enumerate(queryset_iterator(queryset, cls.get_query_limit())):
            delete = not cls.should_index(obj)

            data = {'delete' if delete else 'index': {
                '_index': index_name or cls.get_index_name(),
                '_type': cls.get_type_name(),
                '_id': cls.get_document_id(obj)
            }}
            data.update(cls.get_request_params(obj))

            # bulk operation instructions/details
            tmp.append(data)

            # only append bulk operation data if it's not a delete operation
            if not delete:
                tmp.append(cls.get_document(obj))

            if not i % cls.get_bulk_index_limit():
                es.bulk(tmp)
                tmp = []

        if tmp:
            es.bulk(tmp)

    @classmethod
    def index_add(cls, obj, index_name=''):
        if obj and cls.should_index(obj):
            cls.get_es().index(
                index_name or cls.get_index_name(),
                cls.get_type_name(),
                cls.get_document(obj),
                cls.get_document_id(obj),
                **cls.get_request_params(obj)
            )
            return True
        return False

    @classmethod
    def index_delete(cls, obj, index_name=''):
        if obj:
            try:
                cls.get_es().delete(
                    index_name or cls.get_index_name(),
                    cls.get_type_name(),
                    cls.get_document_id(obj),
                    **cls.get_request_params(obj)
                )
            except TransportError as e:
                if e.status_code != 404:
                    raise
            return True
        return False

    @classmethod
    def index_add_or_delete(cls, obj, index_name=''):
        if obj:
            if cls.should_index(obj):
                return cls.index_add(obj, index_name)
            else:
                return cls.index_delete(obj, index_name)
        return False

    @classmethod
    def save_handler(cls, sender, instance, **kwargs):
        cls.index_add_or_delete(instance)

    @classmethod
    def delete_handler(cls, sender, instance, **kwargs):
        cls.index_delete(instance)



from django.conf import settings


ELASTICSEARCH_SERVER = getattr(settings, 'ELASTICSEARCH_SERVER', ['127.0.0.1:9200', ])
ELASTICSEARCH_CONNECTION_PARAMS = getattr(settings, 'ELASTICSEARCH_CONNECTION_PARAMS', {'hosts': ELASTICSEARCH_SERVER})

# Override this if you want to have a base set of settings for all your indexes. This dictionary
# gets cloned and then updated with custom index-specific from your ELASTICSEARCH_CUSTOM_INDEX_SETTINGS
# Eg. to ensure that all of your indexes have 1 shard and have an edgengram tokenizer/analyzer
# configured
# ELASTICSEARCH_DEFAULT_INDEX_SETTINGS = {
#     "settings" : {
#         "index" : {
#             "number_of_replicas" : 1
#         },
#         "analysis" : {
#             "analyzer" : {
#                 "left" : {
#                     "filter" : [
#                         "standard",
#                         "lowercase",
#                         "stop"
#                     ],
#                     "type" : "custom",
#                     "tokenizer" : "left_tokenizer"
#                 }
#             },
#             "tokenizer" : {
#                 "left_tokenizer" : {
#                     "side" : "front",
#                     "max_gram" : 12,
#                     "type" : "edgeNGram"
#                 }
#             }
#         }
#     }
# }
ELASTICSEARCH_DEFAULT_INDEX_SETTINGS = getattr(settings, 'ELASTICSEARCH_DEFAULT_INDEX_SETTINGS', {})

# Override this in your project settings to define any Elasticsearch-specific index settings.
# Eg.
# ELASTICSEARCH_CUSTOM_INDEX_SETTINGS = {
#     "twitter": {
#         "settings" : {
#             "index" : {
#                 "number_of_shards" : 3,
#             }
#         }
#     },
#     "<your-other-index-name>": {
#         "settings" : {
#             "index" : {
#                 "number_of_shards" : 50,
#                 "number_of_replicas" : 2
#             }
#         }
#     }
# }
ELASTICSEARCH_CUSTOM_INDEX_SETTINGS = getattr(settings, 'ELASTICSEARCH_CUSTOM_INDEX_SETTINGS', {})



import collections
import datetime
import gc
import sys
from django.conf import settings
from django.http import Http404
from django.utils.importlib import import_module
from elasticsearch import Elasticsearch, ElasticsearchException

from . import settings as es_settings


_elasticsearch_indices = collections.defaultdict(lambda: [])


def get_indices(indices=[]):
    if not _elasticsearch_indices:
        type_classes = getattr(settings, 'ELASTICSEARCH_TYPE_CLASSES', ())
        if not type_classes:
            raise Exception('Missing `ELASTICSEARCH_TYPE_CLASSES` in project `settings`.')

        for type_class in type_classes:
            package_name, klass_name = type_class.rsplit('.', 1)
            try:
                package = import_module(package_name)
                klass = getattr(package, klass_name)
            except ImportError:
                sys.stderr.write('Unable to import `{}`.\n'.format(type_class))
                continue
            _elasticsearch_indices[klass.get_index_name()].append(klass)

    if not indices:
        return _elasticsearch_indices
    else:
        result = {}
        for k, v in _elasticsearch_indices.items():
            if k in indices:
                result[k] = v
        return result


def create_aliases(es=None, indices=[]):
    es = es or Elasticsearch(**es_settings.ELASTICSEARCH_CONNECTION_PARAMS)

    current_aliases = es.indices.get_aliases()
    aliases_for_removal = collections.defaultdict(lambda: [])
    for item, tmp in current_aliases.items():
        for iname in list(tmp.get('aliases', {}).keys()):
            aliases_for_removal[iname].append(item)

    actions = []
    for index_alias, index_name in indices:
        for item in aliases_for_removal[index_alias]:
            actions.append({
                'remove': {
                    'index': item,
                    'alias': index_alias
                }
            })
        actions.append({
            'add': {
                'index': index_name,
                'alias': index_alias
            }
        })

    es.indices.update_aliases({'actions': actions})


def create_indices(es=None, indices=[], set_aliases=True):
    es = es or Elasticsearch(**es_settings.ELASTICSEARCH_CONNECTION_PARAMS)

    result = []

    aliases = []
    now = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    for index_alias, type_classes in get_indices(indices).items():
        index_settings = es_settings.ELASTICSEARCH_DEFAULT_INDEX_SETTINGS
        index_settings = recursive_dict_update(
            index_settings,
            es_settings.ELASTICSEARCH_CUSTOM_INDEX_SETTINGS.get(index_alias, {})
        )

        index_name = "{0}-{1}".format(index_alias, now)

        aliases.append((index_alias, index_name))

        type_mappings = {}
        for type_class in type_classes:
            tmp = type_class.get_type_mapping()
            if tmp:
                type_mappings[type_class.get_type_name()] = tmp

            result.append((
                type_class,
                index_alias,
                index_name
            ))

        # if we got any type mappings, put them in the index settings
        if type_mappings:
            index_settings['mappings'] = type_mappings

        es.indices.create(index_name, index_settings)

    if set_aliases:
        create_aliases(es, aliases)

    return result, aliases


def rebuild_indices(es=None, indices=[], set_aliases=True):
    es = es or Elasticsearch(**es_settings.ELASTICSEARCH_CONNECTION_PARAMS)

    created_indices, aliases = create_indices(es, indices, False)

    # kludge to avoid OOM due to Django's query logging
    # db_logger = logging.getLogger('django.db.backends')
    # oldlevel = db_logger.level
    # db_logger.setLevel(logging.ERROR)

    for type_class, index_alias, index_name in created_indices:
        try:
            type_class.bulk_index(es, index_name)
        except NotImplementedError:
            sys.stderr.write('`bulk_index` not implemented on `{}`.\n'.format(type_class.get_index_name()))
            continue

    # return to the norm for db query logging
    # db_logger.setLevel(oldlevel)

    if set_aliases:
        create_aliases(es, aliases)

    return created_indices, aliases


def recursive_dict_update(d, u):
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            r = recursive_dict_update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def queryset_iterator(queryset, chunksize=1000):
    try:
        last_pk = queryset.order_by('-pk')[0].pk
    except IndexError:
        return

    queryset = queryset.order_by('pk')
    pk = queryset[0].pk - 1
    while pk < last_pk:
        for row in queryset.filter(pk__gt=pk)[:chunksize]:
            pk = row.pk
            yield row
        gc.collect()


def get_from_es_or_None(index, type, id, **kwargs):
    es = kwargs.pop('es', Elasticsearch(es_settings.ELASTICSEARCH_SERVER))
    try:
        return es.get(index, id, type, **kwargs)
    except ElasticsearchException:
        return None


def get_from_es_or_404(index, type, id, **kwargs):
    item = get_from_es_or_None(index, type, id, **kwargs)
    if not item:
        raise Http404('No {0} matches the parameters.'.format(type))
    return item



import copy
from datadiff import tools as ddtools
from django import forms
from django.test import TestCase
from elasticsearch import Elasticsearch
from elasticsearch_dsl import Search
from elasticsearch_dsl.result import Response
import mock
from simple_elasticsearch.forms import ESSearchForm, ESSearchProcessor

try:
    # `reload` is not a python3 builtin like python2
    reload
except NameError:
    from imp import reload

from . import settings as es_settings
from .mixins import ElasticsearchIndexMixin
from .models import Blog, BlogPost


class ElasticsearchIndexMixinClass(ElasticsearchIndexMixin):
    pass


class BlogPostSearchForm(ESSearchForm):
    q = forms.CharField()

    def get_index(self):
        return 'blog'

    def get_type(self):
        return 'posts'

    def prepare_query(self):
        return {
            "query": {
                "match_all": {}
            }
        }


class ElasticsearchIndexMixinTestCase(TestCase):

    @property
    def latest_post(self):
        return BlogPost.objects.select_related('blog').latest('id')

    @mock.patch('simple_elasticsearch.mixins.Elasticsearch.delete')
    @mock.patch('simple_elasticsearch.mixins.Elasticsearch.index')
    def setUp(self, mock_index, mock_delete):
        self.blog = Blog.objects.create(
            name='test blog name',
            description='test blog description'
        )

        # hack the return value to ensure we save some BlogPosts here;
        # without this mock, the post_save handler indexing blows up
        # as there is no real ES instance running
        mock_index.return_value = mock_delete.return_value = {}

        post = BlogPost.objects.create(
            blog=self.blog,
            title="DO-NOT-INDEX title",
            slug="DO-NOT-INDEX",
            body="DO-NOT-INDEX body"
        )

        for x in range(1, 10):
            BlogPost.objects.create(
                blog=self.blog,
                title="blog post title {0}".format(x),
                slug="blog-post-title-{0}".format(x),
                body="blog post body {0}".format(x)
            )

    def test__get_es__with_default_settings(self):
        result = BlogPost.get_es()
        self.assertIsInstance(result, Elasticsearch)
        self.assertEqual(result.transport.hosts[0]['host'], '127.0.0.1')
        self.assertEqual(result.transport.hosts[0]['port'], 9200)

    def test__get_es__with_custom_server(self):
        # include a custom class here as the internal `_es` is cached, so can't reuse the
        # `ElasticsearchIndexClassDefaults` global class (see above).
        class ElasticsearchIndexClassCustomSettings(ElasticsearchIndexMixin):
            pass

        with self.settings(ELASTICSEARCH_SERVER=['search.example.com:9201']):
            reload(es_settings)
            result = ElasticsearchIndexClassCustomSettings.get_es()
            self.assertIsInstance(result, Elasticsearch)
            self.assertEqual(result.transport.hosts[0]['host'], 'search.example.com')
            self.assertEqual(result.transport.hosts[0]['port'], 9201)

        reload(es_settings)

    def test__get_es__with_custom_connection_settings(self):
        # include a custom class here as the internal `_es` is cached, so can't reuse the
        # `ElasticsearchIndexClassDefaults` global class (see above).
        class ElasticsearchIndexClassCustomSettings(ElasticsearchIndexMixin):
            pass

        with self.settings(ELASTICSEARCH_CONNECTION_PARAMS={'hosts': ['search2.example.com:9202'], 'sniffer_timeout': 15}):
            reload(es_settings)
            result = ElasticsearchIndexClassCustomSettings.get_es()
            self.assertIsInstance(result, Elasticsearch)
            self.assertEqual(result.transport.hosts[0]['host'], 'search2.example.com')
            self.assertEqual(result.transport.hosts[0]['port'], 9202)
            self.assertEqual(result.transport.sniffer_timeout, 15)
        reload(es_settings)

    @mock.patch('simple_elasticsearch.mixins.ElasticsearchIndexMixin.index_add_or_delete')
    def test__save_handler(self, mock_index_add_or_delete):
        # with a create call
        post = BlogPost.objects.create(
            blog=self.blog,
            title="blog post title foo",
            slug="blog-post-title-foo",
            body="blog post body foo"
        )
        mock_index_add_or_delete.assert_called_with(post)
        mock_index_add_or_delete.reset_mock()

        # with a plain save call
        post.save()
        mock_index_add_or_delete.assert_called_with(post)

    @mock.patch('simple_elasticsearch.mixins.ElasticsearchIndexMixin.index_delete')
    def test__delete_handler(self, mock_index_delete):
        post = self.latest_post
        post.delete()
        mock_index_delete.assert_called_with(post)

    @mock.patch('simple_elasticsearch.mixins.Elasticsearch.index')
    def test__index_add(self, mock_index):
        post = self.latest_post
        mock_index.return_value = {}

        # make sure an invalid object passed in returns False
        result = BlogPost.index_add(None)
        self.assertFalse(result)

        # make sure indexing an item calls Elasticsearch.index() with
        # the correct variables, with normal index name
        result = BlogPost.index_add(post)
        self.assertTrue(result)
        mock_index.assert_called_with('blog', 'posts', BlogPost.get_document(post), post.pk)

        # make sure indexing an item calls Elasticsearch.index() with
        # the correct variables, with non-standard index name
        result = BlogPost.index_add(post, 'foo')
        self.assertTrue(result)
        mock_index.assert_called_with('foo', 'posts', BlogPost.get_document(post), post.pk)

        # this one should not index (return false) because the
        # 'should_index' for this post should make it skip it
        post = BlogPost.objects.get(slug="DO-NOT-INDEX")
        result = BlogPost.index_add(post)
        self.assertFalse(result)

    @mock.patch('simple_elasticsearch.mixins.Elasticsearch.delete')
    def test__index_delete(self, mock_delete):
        post = self.latest_post
        mock_delete.return_value = {
            "acknowledged": True
        }

        # make sure an invalid object passed in returns False
        result = BlogPost.index_delete(None)
        self.assertFalse(result)

        # make sure deleting an item calls Elasticsearch.delete() with
        # the correct variables, with normal index name
        result = BlogPost.index_delete(post)
        self.assertTrue(result)
        mock_delete.assert_called_with('blog', 'posts', post.pk)

        # make sure deleting an item calls Elasticsearch.delete() with
        # the correct variables, with non-standard index name
        result = BlogPost.index_delete(post, 'foo')
        self.assertTrue(result)
        mock_delete.assert_called_with('foo', 'posts', post.pk)

    @mock.patch('simple_elasticsearch.mixins.ElasticsearchIndexMixin.index_add')
    @mock.patch('simple_elasticsearch.mixins.ElasticsearchIndexMixin.index_delete')
    def test__index_add_or_delete(self, mock_index_delete, mock_index_add):
        # invalid object passed in, should return False
        result = BlogPost.index_add_or_delete(None)
        self.assertFalse(result)

        # this one should not index (return false) because the
        # `should_index` for this post should make it skip it;
        # `index_delete` should get called
        mock_index_delete.return_value = True
        post = BlogPost.objects.get(slug="DO-NOT-INDEX")

        result = BlogPost.index_add_or_delete(post)
        self.assertTrue(result)
        mock_index_delete.assert_called_with(post, '')

        result = BlogPost.index_add_or_delete(post, 'foo')
        self.assertTrue(result)
        mock_index_delete.assert_called_with(post, 'foo')

        # `index_add` call results below
        mock_index_add.return_value = True
        post = self.latest_post

        result = BlogPost.index_add_or_delete(post)
        self.assertTrue(result)
        mock_index_add.assert_called_with(post, '')

        result = BlogPost.index_add_or_delete(post, 'foo')
        self.assertTrue(result)
        mock_index_add.assert_called_with(post, 'foo')

    def test__get_index_name(self):
        self.assertEqual(BlogPost.get_index_name(), 'blog')

    def test__get_type_name(self):
        self.assertEqual(BlogPost.get_type_name(), 'posts')

    def test__get_queryset(self):
        queryset = BlogPost.objects.all().select_related('blog').order_by('pk')
        self.assertEqual(list(BlogPost.get_queryset().order_by('pk')), list(queryset))

    def test__get_index_name_notimplemented(self):
        with self.assertRaises(NotImplementedError):
            ElasticsearchIndexMixinClass.get_index_name()

    def test__get_type_name_notimplemented(self):
        with self.assertRaises(NotImplementedError):
            ElasticsearchIndexMixinClass.get_type_name()

    def test__get_queryset_notimplemented(self):
        with self.assertRaises(NotImplementedError):
            ElasticsearchIndexMixinClass.get_queryset()

    def test__get_type_mapping(self):
        mapping = {
            "properties": {
                "created_at": {
                    "type": "date",
                    "format": "dateOptionalTime"
                },
                "title": {
                    "type": "string"
                },
                "body": {
                    "type": "string"
                },
                "slug": {
                    "type": "string"
                },
                "blog": {
                    "properties": {
                        "id": {
                            "type": "long"
                        },
                        "name": {
                            "type": "string"
                        },
                        "description": {
                            "type": "string"
                        }
                    }
                }
            }
        }
        self.assertEqual(BlogPost.get_type_mapping(), mapping)

    def test__get_type_mapping_notimplemented(self):
        self.assertEqual(ElasticsearchIndexMixinClass.get_type_mapping(), {})

    def test__get_request_params(self):
        post = self.latest_post
        # TODO: implement the method to test it works properly
        self.assertEqual(BlogPost.get_request_params(post), {})

    def test__get_request_params_notimplemented(self):
        self.assertEqual(ElasticsearchIndexMixinClass.get_request_params(1), {})

    def test__get_bulk_index_limit(self):
        self.assertTrue(str(BlogPost.get_bulk_index_limit()).isdigit())

    def test__get_query_limit(self):
        self.assertTrue(str(BlogPost.get_query_limit()).isdigit())

    def test__get_document_id(self):
        post = self.latest_post
        result = BlogPost.get_document_id(post)
        self.assertEqual(result, post.pk)

    def test__get_document(self):
        post = self.latest_post
        result = BlogPost.get_document(post)
        self.assertEqual(result, {
            'title': post.title,
            'slug': post.slug,
            'blog': {
                'id': post.blog.pk,
                'description': post.blog.description,
                'name': post.blog.name
            },
            'created_at': post.created_at,
            'body': post.body
        })

    def test__get_document_notimplemented(self):
        with self.assertRaises(NotImplementedError):
            ElasticsearchIndexMixinClass.get_document(1)

    @mock.patch('simple_elasticsearch.mixins.Elasticsearch.index')
    def test__should_index(self, mock_index):
        post = self.latest_post
        self.assertTrue(BlogPost.should_index(post))

        mock_index.return_value = {}
        post = BlogPost.objects.get(slug="DO-NOT-INDEX")
        self.assertFalse(BlogPost.should_index(post))

    def test__should_index_notimplemented(self):
        self.assertTrue(ElasticsearchIndexMixinClass.should_index(1))

    @mock.patch('simple_elasticsearch.mixins.queryset_iterator')
    def test__bulk_index_queryset(self, mock_queryset_iterator):
        queryset = BlogPost.get_queryset().exclude(slug='DO-NOT-INDEX')
        BlogPost.bulk_index(queryset=queryset)
        mock_queryset_iterator.assert_called_with(queryset, BlogPost.get_query_limit())

        mock_queryset_iterator.reset_mock()

        queryset = BlogPost.get_queryset()
        BlogPost.bulk_index()
        # to compare QuerySets, they must first be converted to lists.
        self.assertEqual(list(mock_queryset_iterator.call_args[0][0]), list(queryset))

    @mock.patch('simple_elasticsearch.models.BlogPost.get_document')
    @mock.patch('simple_elasticsearch.models.BlogPost.should_index')
    @mock.patch('simple_elasticsearch.mixins.Elasticsearch.bulk')
    def test__bulk_index_should_index(self, mock_bulk, mock_should_index, mock_get_document):
        # hack the return value to ensure we save some BlogPosts here;
        # without this mock, the post_save handler indexing blows up
        # as there is no real ES instance running
        mock_bulk.return_value = {}

        queryset_count = BlogPost.get_queryset().count()
        BlogPost.bulk_index()
        self.assertTrue(mock_should_index.call_count == queryset_count)

    @mock.patch('simple_elasticsearch.models.BlogPost.get_document')
    @mock.patch('simple_elasticsearch.mixins.Elasticsearch.bulk')
    def test__bulk_index_get_document(self, mock_bulk, mock_get_document):
        mock_bulk.return_value = mock_get_document.return_value = {}

        queryset_count = BlogPost.get_queryset().count()
        BlogPost.bulk_index()

        # One of the items is not meant to be indexed (slug='DO-NOT-INDEX'), so the
        # get_document function will get called one less time due to this.
        self.assertTrue(mock_get_document.call_count == (queryset_count - 1))

        # figure out how many times es.bulk() should get called in the
        # .bulk_index() method and verify it's the same
        bulk_times = int(queryset_count / BlogPost.get_bulk_index_limit()) + 1
        self.assertTrue(mock_bulk.call_count == bulk_times)


class ESSearchFormTestCase(TestCase):

    def setUp(self):
        self.query = {'q': 'python'}
        self.form = BlogPostSearchForm(self.query)
        self.form.is_valid()

    def test__form_get_index(self):
        self.assertEqual(self.form.get_index(), 'blog')

    def test__form_get_type(self):
        self.assertEqual(self.form.get_type(), 'posts')

    def test__form_query_params(self):
        self.assertEqual(self.form.query_params, {})

        query_params = {'test': 'foo'}
        form = BlogPostSearchForm(query_params=query_params)
        self.assertEqual(form.query_params, query_params)

    @mock.patch('simple_elasticsearch.forms.ESSearchProcessor.search')
    @mock.patch('simple_elasticsearch.forms.ESSearchProcessor.add_search')
    @mock.patch('simple_elasticsearch.forms.ESSearchProcessor.__init__')
    def test__form_es(self, mock_esp_init, mock_esp_add_search, mock_esp_search):
        # __init__ methods always return None
        mock_esp_init.return_value = None

        # this allows the form.search() method to complete without
        # exceptions being raised
        mock_esp_search.return_value = [Response({})]

        # by default, the form has no internal Elasticsearch object
        self.assertEqual(self.form.es, None)

        # on search(), ESSearchProcessor() should be initialized with
        # a None Elasticsearch object (the form's)
        self.form.search()
        mock_esp_init.assert_called_with(None)
        mock_esp_init.reset()

        # here, we're setting the form's internal Elasticsearch
        # object, so the above tests should have the opposite
        # result
        form = BlogPostSearchForm(es=Elasticsearch(['127.0.0.2:9201']))
        self.assertIsInstance(form.es, Elasticsearch)
        self.assertEqual(form.es.transport.hosts[0]['host'], '127.0.0.2')
        self.assertEqual(form.es.transport.hosts[0]['port'], 9201)

        form.search()
        mock_esp_init.assert_called_with(form.es)

    def test__form_data_validation(self):
        form = BlogPostSearchForm({})
        self.assertFalse(form.is_valid())

        form = BlogPostSearchForm({'q': ''})
        self.assertFalse(form.is_valid())

        form = BlogPostSearchForm({'q': 'foo'})
        self.assertTrue(form.is_valid())

    @mock.patch('simple_elasticsearch.forms.ESSearchProcessor.search')
    @mock.patch('simple_elasticsearch.forms.ESSearchProcessor.add_search')
    def test__form_search(self, mock_esp_add_search, mock_esp_search):
        mock_esp_search.return_value = [Response({})]
        self.form.search()
        mock_esp_add_search.assert_called_with(self.form, 1, 20)
        mock_esp_add_search.reset()

        self.form.search(5, 50)
        mock_esp_add_search.assert_called_with(self.form, 5, 50)


class ESSearchProcessorTestCase(TestCase):

    def setUp(self):
        self.query = {'q': 'python'}
        self.form = BlogPostSearchForm(self.query)
        self.form.is_valid()

    def test__esp_reset(self):
        esp = ESSearchProcessor()

        self.assertTrue(len(esp.bulk_search_data) == 0)
        self.assertTrue(len(esp.page_ranges) == 0)

        esp.add_search(self.form)

        self.assertFalse(len(esp.bulk_search_data) == 0)
        self.assertFalse(len(esp.page_ranges) == 0)

        esp.reset()

        self.assertTrue(len(esp.bulk_search_data) == 0)
        self.assertTrue(len(esp.page_ranges) == 0)

    def test__esp_add_query_dict(self):
        esp = ESSearchProcessor()

        page = 1
        page_size = 20

        query = {
            "query": {
                "match": {
                    "_all": "foobar"
                }
            }
        }

        # ESSearchProcessor internally sets the from/size parameters
        # on the query; we need to compare with those values included
        query_with_size = query.copy()
        query_with_size.update({
            'from': (page - 1) * page_size,
            'size': page_size
        })

        esp.add_search(query.copy())
        ddtools.assert_equal(esp.bulk_search_data[0], {})
        ddtools.assert_equal(esp.bulk_search_data[1], query_with_size)

        esp.reset()
        esp.add_search(query.copy(), index='blog')
        ddtools.assert_equal(esp.bulk_search_data[0], {'index': 'blog'})
        ddtools.assert_equal(esp.bulk_search_data[1], query_with_size)

        esp.reset()
        esp.add_search(query.copy(), index='blog', doc_type='posts')
        ddtools.assert_equal(esp.bulk_search_data[0], {'index': 'blog', 'type': 'posts'})
        ddtools.assert_equal(esp.bulk_search_data[1], query_with_size)

    def test__esp_add_query_form(self):
        esp = ESSearchProcessor()

        page = 1
        page_size = 20

        query = self.form.prepare_query()

        # ESSearchProcessor internally sets the from/size parameters
        # on the query; we need to compare with those values included
        query_with_size = query.copy()
        query_with_size.update({
            'from': (page - 1) * page_size,
            'size': page_size
        })

        esp.add_search(self.form)
        ddtools.assert_equal(esp.bulk_search_data[0], {'index': 'blog', 'type': 'posts'})
        ddtools.assert_equal(esp.bulk_search_data[1], query_with_size)

    def test__esp_add_query_dslquery(self):
        page = 1
        page_size = 20

        query = {
            "query": {
                "match": {
                    "_all": "foobar"
                }
            }
        }

        s = Search.from_dict(query.copy())

        # ESSearchProcessor internally sets the from/size parameters
        # on the query; we need to compare with those values included
        query_with_size = query.copy()
        query_with_size.update({
            'from': (page - 1) * page_size,
            'size': page_size
        })

        esp = ESSearchProcessor()
        esp.add_search(s)
        ddtools.assert_equal(esp.bulk_search_data[0], {})
        ddtools.assert_equal(esp.bulk_search_data[1], query_with_size)

        esp.reset()
        esp.add_search(s, index='blog')
        ddtools.assert_equal(esp.bulk_search_data[0], {'index': 'blog'})
        ddtools.assert_equal(esp.bulk_search_data[1], query_with_size)

        esp.reset()
        esp.add_search(s, index='blog', doc_type='posts')
        ddtools.assert_equal(esp.bulk_search_data[0], {'index': 'blog', 'type': 'posts'})
        ddtools.assert_equal(esp.bulk_search_data[1], query_with_size)

        s = s.index('blog').params(routing='id')

        esp.reset()
        esp.add_search(s)
        ddtools.assert_equal(esp.bulk_search_data[0], {'index': ['blog'], 'routing': 'id'})
        ddtools.assert_equal(esp.bulk_search_data[1], query_with_size)

        s = s.doc_type('posts')

        esp.reset()
        esp.add_search(s)
        ddtools.assert_equal(esp.bulk_search_data[0], {'index': ['blog'], 'type': ['posts'], 'routing': 'id'})
        ddtools.assert_equal(esp.bulk_search_data[1], query_with_size)

    @mock.patch('simple_elasticsearch.forms.Elasticsearch.msearch')
    def test__esp_search(self, mock_msearch):
        mock_msearch.return_value = {
            "responses": [
                {
                    "hits": {
                        "hits": []
                    }
                }
            ]
        }

        esp = ESSearchProcessor()
        esp.add_search({}, index='blog', doc_type='posts')

        bulk_data = copy.deepcopy(esp.bulk_search_data)
        ddtools.assert_equal(bulk_data, [{'index': 'blog', 'type': 'posts'}, {'from': 0, 'size': 20}])

        responses = esp.search()
        mock_msearch.assert_called_with(bulk_data)

        # ensure that our hack to get size and from into the hit
        # data works
        self.assertTrue('size' in responses[0].get('hits'))
        self.assertTrue('from' in responses[0].get('hits'))

        # ensure that the bulk data gets reset
        self.assertEqual(len(esp.bulk_search_data), 0)



import sys
from optparse import make_option
from django.core.management.base import BaseCommand, CommandError

from ...utils import get_indices, create_indices, rebuild_indices


class ESCommandError(CommandError):
    pass


class Command(BaseCommand):
    help = ''
    option_list = BaseCommand.option_list + (
        make_option(
            '--list',
            action='store_true',
            dest='list',
            default=False,
        ),
        make_option(
            '--initialize',
            action='store_true',
            dest='initialize',
            default=False,
        ),
        make_option(
            '--rebuild',
            action='store_true',
            dest='rebuild',
            default=False,
        ),
        make_option(
            '--no_input',
            action='store_true',
            dest='no_input',
            default=False,
        ),
        make_option(
            '--indexes',
            action='store',
            dest='indexes',
            default='',
        )
    )

    def handle(self, *args, **options):
        no_input = options.get('no_input')

        requested_indexes = options.get('indexes', '') or []
        if requested_indexes:
            requested_indexes = requested_indexes.split(',')

        if options.get('list'):
            self.subcommand_list()
        elif options.get('initialize'):
            self.subcommand_initialize(requested_indexes, no_input)
        elif options.get('rebuild'):
            self.subcommand_rebuild(requested_indexes, no_input)

    def subcommand_list(self):
        print("Available ES indexes:")
        for index_name, type_classes in get_indices().items():
            print(" - index '{0}':".format(index_name))
            for type_class in type_classes:
                print("  - type '{0}'".format(type_class.get_type_name()))

    def subcommand_initialize(self, indexes=None, no_input=False):
        input = 'y' if no_input else ''
        while input != 'y':
            input = input('Are you sure you want to initialize {0} index(es)? [y/N]: '.format('the ' + ', '.join(indexes) if indexes else '**ALL**')).lower()
            if input == 'n':
                break

        if input == 'y':
            sys.stdout.write("Creating ES indexes: ")
            results, aliases = create_indices(indices=indexes)
            sys.stdout.write("complete.\n")
            for alias, index in aliases:
                print("'{0}' aliased to '{1}'".format(alias, index))

    def subcommand_rebuild(self, indexes, no_input=False):
        input = 'y' if no_input else ''
        while input != 'y':
            input = input('Are you sure you want to rebuild {0} index(es)? [y/N]: '.format('the ' + ', '.join(indexes) if indexes else '**ALL**')).lower()
            if input == 'n':
                break

        if input == 'y':
            sys.stdout.write("Rebuilding ES indexes: ")
            results, aliases = rebuild_indices(indices=indexes)
            sys.stdout.write("complete.\n")
            for alias, index in aliases:
                print("'{0}' rebuilt and aliased to '{1}'".format(alias, index))

        # TODO: need to offer choice to delete old de-aliased indexes
        # while input != 'y':
        #     input = raw_input('Are you sure you want to rebuild {0} index(es)? [y/N]: '.format('the ' + ', '.join(indexes) if indexes else '**ALL**')).lower()
        #     if input == 'n':
        #         break
        #
        # if input == 'y':
        #     sys.stdout.write("Rebuilding ES indexes: ")
        #     results, aliases = rebuild_indices(indexes)
        #     sys.stdout.write("complete.\n")
        #     for alias, index in aliases:
        #         print "'{0}' rebuilt and aliased to '{1}'".format(alias, index)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import simple_elasticsearch

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

requirements = [
    'elasticsearch>=1.0.0,<2.0.0',
    'elasticsearch-dsl>=0.0.2'
]

test_requirements = [
    'datadiff>=1.1.6',
    'mock>=1.0.1'
]

setup(
    name='django-simple-elasticsearch',
    version=simple_elasticsearch.__version__,
    description='Simple ElasticSearch indexing integration for Django.',
    long_description=readme + '\n\n' + history,
    author='James Addison',
    author_email='code@scottisheyes.com',
    url='https://github.com/jaddison/django-simple-elasticsearch',
    packages=[
        'simple_elasticsearch',
        'simple_elasticsearch.management',
        'simple_elasticsearch.management.commands',
    ],
    package_dir={'django-simple-elasticsearch':
                 'django-simple-elasticsearch'},
    include_package_data=True,
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords='django simple elasticsearch search indexing haystack',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'Programming Language :: Python',
        'Framework :: Django',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    test_suite='runtests.runtests',
    tests_require=test_requirements
)
